"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.GenerateTypesCommand = void 0;
const loaders_1 = require("../loaders");
const chalk_1 = __importDefault(require("chalk"));
const utils_1 = require("../utils");
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const process_1 = __importDefault(require("process"));
const chokidar_1 = __importDefault(require("chokidar"));
const utils_2 = require("../utils");
const fs_extra_1 = require("fs-extra");
const import_1 = require("../utils/import");
class GenerateTypesCommand {
    constructor() {
        this.command = 'generate-types';
        this.describe = 'Generate types for translations. Supports json and yaml files.';
    }
    builder(args) {
        return args
            .option('debounce', {
            alias: 'd',
            type: 'number',
            describe: 'Debounce time in ms',
            default: 200,
            demandOption: false,
        })
            .option('optionsFile', {
            alias: 'opt',
            type: 'string',
            describe: 'Options file path',
            demandOption: false,
        })
            .option('watch', {
            alias: 'w',
            type: 'boolean',
            describe: 'Watch for changes and generate types',
            default: false,
            demandOption: false,
        })
            .option('typesOutputPath', {
            alias: 'o',
            type: 'string',
            describe: 'Path to output types file',
            default: 'src/generated/i18n.generated.ts',
            demandOption: false,
        })
            .option('loaderType', {
            alias: 't',
            type: 'string',
            array: true,
            options: ['json', 'yaml'],
            describe: 'Loader type',
            demandOption: false,
            default: [],
        })
            .option('translationsPath', {
            alias: 'p',
            type: 'string',
            describe: 'Path to translations',
            array: true,
            default: [],
            demandOption: false,
        });
    }
    async handler(args) {
        const { packageConfig = {}, packageJsonFilePath } = (await getPackageConfig()) || {};
        packageConfig.i18n = packageConfig.i18n || {};
        if (!args.typesOutputPath && packageConfig.i18n.typesOutputPath) {
            args.typesOutputPath = packageConfig.i18n.typesOutputPath;
        }
        if (args.optionsFile) {
            args.optionsFile = path_1.default.resolve(process_1.default.cwd(), args.optionsFile);
        }
        if (!args.optionsFile && packageConfig.i18n.optionsFile) {
            const packageJsonFolder = path_1.default.dirname(packageJsonFilePath);
            args.optionsFile = path_1.default.join(packageJsonFolder, packageConfig.i18n.optionsFile);
        }
        if (!args.typesOutputPath) {
            console.log(chalk_1.default.red(`Error: typesOutputPath is not defined. Please provide a path to output types file, in params or in package.json`));
            process_1.default.exit(1);
        }
        args.translationsPath = sanitizePaths(args.translationsPath);
        validateInputParams(args);
        validatePathsNotEmbeddedInEachOther(args.translationsPath);
        const optionsFromFile = await validateAndGetOptionsFile(args.optionsFile);
        const loaders = args.loaderType.map((loaderType, index) => {
            const path = args.translationsPath[index];
            validatePath(path, loaderType, index);
            return {
                path,
                loader: getLoaderByType(loaderType, path),
            };
        });
        (optionsFromFile?.loaders || []).forEach((loader) => {
            const p = loader?.options?.path;
            loaders.push({
                path: p ? p : sanitizePath(p),
                loader: loader,
            });
        });
        const translationsWithPaths = await loadTranslations(loaders);
        let hasError = false;
        const translationsMapped = translationsWithPaths.map(({ translations, error, path }) => {
            if (error) {
                console.log(chalk_1.default.red(`Error while loading translations from ${path}: ${error.message}`));
                hasError = true;
            }
            return translations;
        });
        if (!hasError) {
            const mergedTranslations = reduceTranslations(translationsMapped);
            await generateAndSaveTypes(mergedTranslations, args);
        }
        else if (!args.watch) {
            process_1.default.exit(1);
        }
        if (args.watch) {
            console.log(chalk_1.default.green(`Listening for changes in ${args.translationsPath.join(', ')}...`));
            if (this.fsWatcher === undefined) {
                this.fsWatcher = await listenForChanges(loaders, translationsWithPaths, args);
            }
        }
        else {
            process_1.default.exit(0);
        }
    }
    async stopWatcher() {
        if (this.fsWatcher) {
            await this.fsWatcher.close();
        }
    }
}
exports.GenerateTypesCommand = GenerateTypesCommand;
function validatePathsNotEmbeddedInEachOther(paths) {
    for (let i = 0; i < paths.length; i++) {
        const pathToCheck = paths[i];
        for (let j = 0; j < paths.length; j++) {
            const pathToCompare = paths[j];
            if (j !== i && pathToCheck.startsWith(pathToCompare)) {
                console.log(chalk_1.default.red(`Path ${pathToCheck} is embedded in ${pathToCompare}. This is not supported.`));
                process_1.default.exit(1);
            }
        }
    }
}
function listenForChanges(loadersWithPaths, translationsWithPaths, args) {
    const allPaths = loadersWithPaths.map(({ path }) => path);
    const loadersByPath = loadersWithPaths.reduce((acc, { path, loader }) => {
        acc[path] = loader;
        return acc;
    }, {});
    return new Promise((resolve, reject) => {
        const fsWatcher = chokidar_1.default
            .watch(allPaths, {
            ignoreInitial: true,
        })
            .on('ready', () => {
            resolve(fsWatcher);
        })
            .on('error', (error) => {
            console.log(chalk_1.default.red(`Error while watching files: ${error.message}`));
            reject(error);
        })
            .on('all', customDebounce(handleFileChangeEvents(allPaths, loadersByPath, translationsWithPaths, args), args.debounce));
    });
}
function sanitizePath(path) {
    const newPath = path.endsWith('/') ? path : `${path}/`;
    return newPath.startsWith('./') ? newPath.slice(2) : newPath;
}
function sanitizePaths(paths) {
    return paths.map((path) => {
        return sanitizePath(path);
    });
}
function handleFileChangeEvents(listenToPaths, loadersByPath, translationsWithPaths, args) {
    return async (events, paths) => {
        console.log(chalk_1.default.blue(`Change detected`));
        console.log(chalk_1.default.green(`${events.map((e, idx) => `\t${e} - ${paths[idx]}`).join('\n')}`));
        console.log(chalk_1.default.blue(`Re-generating types...`));
        const uniquePaths = new Set();
        for (const changePath of paths) {
            const foundPath = listenToPaths.find((path) => changePath.startsWith(path));
            uniquePaths.add(foundPath);
            if (uniquePaths.size === paths.length) {
                break;
            }
        }
        let hasError = false;
        for (const path of uniquePaths) {
            const loader = loadersByPath[path];
            try {
                const translation = (await loader.load());
                translationsWithPaths.forEach((translationWithPath) => {
                    if (translationWithPath.path === path) {
                        translationWithPath.translations = translation;
                    }
                });
            }
            catch (e) {
                hasError = true;
                console.log(chalk_1.default.red(`Error while loading translations from ${path}. Error: ${e.message}`));
            }
        }
        if (hasError) {
            console.log(chalk_1.default.red(`Waiting for changes to generate proper types`));
            return;
        }
        const mergedTranslations = reduceTranslations(translationsWithPaths.map(({ translations }) => translations));
        await generateAndSaveTypes(mergedTranslations, args);
    };
}
async function generateAndSaveTypes(translations, args) {
    const object = Object.keys(translations).reduce((result, key) => (0, utils_1.mergeDeep)(result, translations[key]), {});
    const rawContent = await (0, utils_2.createTypesFile)(object);
    const outputFile = (0, utils_2.annotateSourceCode)(rawContent);
    fs_1.default.mkdirSync(path_1.default.dirname(args.typesOutputPath), {
        recursive: true,
    });
    let currentFileContent = null;
    try {
        currentFileContent = fs_1.default.readFileSync(args.typesOutputPath, 'utf8');
    }
    catch (err) { }
    if (currentFileContent != outputFile) {
        fs_1.default.writeFileSync(args.typesOutputPath, outputFile);
        console.log(`
        ${chalk_1.default.green(`Types generated and saved to: ${args.typesOutputPath}`)}
      `);
    }
    else {
        console.log(`
        ${chalk_1.default.yellow('No changes generated in a result output type file.')}
      `);
    }
}
function customDebounce(func, wait) {
    let args = [];
    let timeoutId;
    return function (...rest) {
        if (func.length) {
            for (let i = 0; i < func.length; i++) {
                if (!args[i]) {
                    args[i] = [];
                }
                args[i].push(rest[i]);
            }
        }
        else {
            args.push(...rest);
        }
        clearTimeout(timeoutId);
        timeoutId = setTimeout(function () {
            func.apply(this, args);
            args = [];
        }, wait);
    };
}
async function loadTranslations(loaders) {
    const loadedTranslations = (await Promise.all(loaders.map(({ loader }) => loader.load().catch((e) => e))));
    return loadedTranslations.map((translations, index) => {
        const error = translations instanceof Error;
        return {
            translations: error ? null : translations,
            error: error ? translations : null,
            path: loaders[index].path,
        };
    });
}
async function validateAndGetOptionsFile(optionsFile) {
    if (optionsFile) {
        let optionsFileExport;
        try {
            optionsFileExport = await (0, import_1.importOrRequireFile)(optionsFile);
        }
        catch (err) {
            throw new Error(`Unable to open file: "${optionsFile}". ${err.message}`);
        }
        if (!optionsFileExport || typeof optionsFileExport !== 'object') {
            throw new Error(`Given options file must contain export of a I18nOptions instance`);
        }
        const optionsExported = [];
        for (const key in optionsFileExport) {
            const options = optionsFileExport[key];
            if (options.loaders) {
                optionsExported.push(options);
            }
        }
        if (optionsExported.length === 0) {
            throw new Error(`Given options file must contain export of a I18nOptions`);
        }
        if (optionsExported.length > 1) {
            throw new Error(`Given options file must contain only one export of I18nOptions`);
        }
        return optionsExported[0];
    }
}
function validateInputParams(args) {
    if (args.loaderType.length !== args.translationsPath.length) {
        console.log(chalk_1.default.red(`Error: translationsPath and loaderType must have the same number of elements.
           You provided ${args.loaderType.length} loader types and ${args.translationsPath.length} paths`));
        process_1.default.exit(1);
    }
    if ((args.loaderType.length === 0 || args.loaderType.length === 0) &&
        !args.optionsFile) {
        console.log(chalk_1.default.red(`Error: you must provide at least one loader type or options file`));
        process_1.default.exit(1);
    }
}
function validatePath(path, loaderType, index) {
    if (path === undefined) {
        console.log(chalk_1.default.red(`Error: translationsPath is not defined for loader type ${loaderType},
                 please provide a path to translations, index ${index}`));
        process_1.default.exit(1);
    }
}
async function getPackageConfig(basePath = process_1.default.cwd()) {
    const packageJsonFilePath = `${basePath}/package.json`;
    if (await (0, fs_extra_1.pathExists)(packageJsonFilePath)) {
        try {
            const packageConfig = await require(packageJsonFilePath);
            return {
                packageJsonFilePath,
                packageConfig,
            };
        }
        catch {
            return undefined;
        }
    }
    const parentFolder = await (0, fs_extra_1.realpath)(`${basePath}/..`);
    if (basePath === parentFolder) {
        return undefined;
    }
    return this.getPackageConfig(parentFolder);
}
function getLoaderByType(loaderType, path) {
    switch (loaderType) {
        case 'json':
            return new loaders_1.I18nJsonLoader({
                path,
            });
        case 'yaml':
            return new loaders_1.I18nYamlLoader({
                path,
            });
        default:
            console.log(chalk_1.default.red(`Error: loader type ${loaderType} is not supported`));
            process_1.default.exit(1);
    }
}
function reduceTranslations(translations) {
    return translations.reduce((acc, t) => (0, utils_1.mergeTranslations)(acc, t), {});
}
//# sourceMappingURL=generate-types.command.js.map