"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.I18nAbstractFileLoader = void 0;
const i18n_loader_1 = require("./i18n.loader");
const path_1 = __importDefault(require("path"));
const promises_1 = require("fs/promises");
const file_1 = require("../utils/file");
const i18n_error_1 = require("../i18n.error");
class I18nAbstractFileLoader extends i18n_loader_1.I18nLoader {
    constructor(options) {
        super(options);
        this.options = this.sanitizeOptions(options);
    }
    async languages() {
        return this.parseLanguages();
    }
    async load() {
        return this.parseTranslations();
    }
    async parseTranslations() {
        const i18nPath = path_1.default.normalize(this.options.path + path_1.default.sep);
        const translations = {};
        if (!(await (0, file_1.exists)(i18nPath))) {
            throw new i18n_error_1.I18nError(`i18n path (${i18nPath}) cannot be found`);
        }
        if (!this.options.filePattern.match(/\*\.[A-z]+/)) {
            throw new i18n_error_1.I18nError(`filePattern should be formatted like: *.json, *.txt, *.custom etc`);
        }
        const languages = await this.parseLanguages();
        const pattern = new RegExp('.' + this.options.filePattern.replace('.', '.'));
        const files = await [
            ...languages.map((l) => path_1.default.join(i18nPath, l)),
            i18nPath,
        ].reduce(async (f, p) => {
            (await f).push(...(await (0, file_1.getFiles)(p, pattern, this.options.includeSubfolders)));
            return f;
        }, Promise.resolve([]));
        for (const file of files) {
            let global = false;
            const pathParts = path_1.default
                .dirname(path_1.default.relative(i18nPath, file))
                .split(path_1.default.sep);
            const key = pathParts[0];
            if (key === '.') {
                global = true;
            }
            const source = await (0, promises_1.readFile)(file, 'utf8');
            const data = this.formatData(source, file);
            const prefix = [...pathParts.slice(1), path_1.default.basename(file).split('.')[0]];
            for (const property of Object.keys(data)) {
                [...(global ? languages : [key])].forEach((lang) => {
                    translations[lang] = translations[lang] ? translations[lang] : {};
                    if (global) {
                        translations[lang][property] = data[property];
                    }
                    else {
                        this.assignPrefixedTranslation(translations[lang], prefix, property, data[property]);
                    }
                });
            }
        }
        return translations;
    }
    assignPrefixedTranslation(translations, prefix, property, value) {
        if (prefix.length) {
            translations[prefix[0]] = translations[prefix[0]]
                ? translations[prefix[0]]
                : {};
            this.assignPrefixedTranslation(translations[prefix[0]], prefix.slice(1), property, value);
        }
        else {
            translations[property] = value;
        }
    }
    async parseLanguages() {
        const i18nPath = path_1.default.normalize(this.options.path + path_1.default.sep);
        return (await (0, file_1.getDirectories)(i18nPath)).map((dir) => path_1.default.relative(i18nPath, dir));
    }
    sanitizeOptions(options) {
        options = { ...this.getDefaultOptions(), ...options };
        options.path = path_1.default.normalize(options.path + path_1.default.sep);
        if (!options.filePattern.startsWith('*.')) {
            options.filePattern = '*.' + options.filePattern;
        }
        return options;
    }
}
exports.I18nAbstractFileLoader = I18nAbstractFileLoader;
//# sourceMappingURL=i18n.abstract-file.loader.js.map